#!/usr/bin/perl
##############################################################################
##                                                                           #
##   Scripts de configuracion para Lihuen 4 GNU/Linux                        #
##   Copyright (C) 2011 Fernando López, Atilio Gonzalez, Sofía Martin,       #
##   Joaquín Bogado, Aldo Vizcaino, Damián Mel, Andrea Gómez del Monáco      #
##   Copyright (C) 2009-2010 Fernando López, Joaquín Bogado, Sofía Martin,   #
##   Raul Benencia, Atilio Gonzalez, Aldo Vizcaino                           #
##   soportelihuen [en] linti.unlp.edu.ar                                    #
##   Copyright (C) 2008  Fernando López, Joaquín Bogado, Sofía Martin,       #
##   Raul Benencia, Atilio Gonzalez, Aldo Vizcaino					         #
##   Copyright (C) 2007  Fernando López, Joaquín Bogado y Sofía Martin       #
##   Copyright (C) 2006  Fernando López y Andrea Pazo                        #
##                                                                           #
##   This program is free software: you can redistribute it and/or modify    #
##   it under the terms of the GNU General Public License as published by    #
##   the Free Software Foundation, either version 3 of the License, or       #
##   (at your option) any later version.                                     #
##                                                                           #
##   This program is distributed in the hope that it will be useful,         #
##   but WITHOUT ANY WARRANTY; without even the implied warranty of          #
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
##   GNU General Public License for more details.                            #
##                                                                           #
##   You should have received a copy of the GNU General Public License       #
##   along with this program.  If not, see <http://www.gnu.org/licenses/>    #
##                                                                           #
##############################################################################
#Requiere udevadm
###########Requiere fdisk, /etc/fstab, mkdir-rmdir, mount-umount y privilegios de root

use strict;
use warnings;

my $basename = '.+/([[:alnum:]]+$)';

my %dumpAndCheck = (
	"no_nativo" => "0\t0",
	"nativo" => "0\t0",
	"root" => "0\t1",
	"swap" => "0\t0"
);

my %clasificacion = (
	"ext2" => "nativo", "ext3" => "nativo", "ext4" => "nativo", "btrfs" => "nativo",
	"reiserfs" => "nativo", "reiser4" => "nativo",
	"vfat" => "no_nativo", "fat" => "no_nativo", "udf,iso9660" => "no_nativo",
	"ntfs" => "no_nativo",
	"swap" => "swap"
);

my %opciones = (
	"nativo" => "defaults",
	"vfat" => "fmask=666,gid=500,dmask=777,umask=0002,noexec,nodev,rw,auto,nouser",
	"ntfs" => "force,gid=500,rw,auto,noexec,nodev,nouser,fmask=666,dmask=777,umask=0002,nls=utf8",
	"swap" => "sw", "root" => "errors=remount-ro", "udf,iso9660" => "users,noauto",
	"fatfloppy" => "rw,user,noauto"
);

sub opciones($$){
	my ($prop, $root) = @_;
    my ($dc, $clas, $ops) = ["", "", ""];
    # Se monta en /
    return $opciones{'root'} . "\t" . $dumpAndCheck{'root'} if ($root);

    # Si es un disquete (asumimos vfat)
	if (es_floppy($prop)){
    	return $opciones{'fatfloppy'} . "\t" . $dumpAndCheck{'no_nativo'};
	}

	if (es_cdrom($prop)){
		return $opciones{'udf,iso9660'} . "\t" . $dumpAndCheck{'no_nativo'};
	}

	$clas = $clasificacion{$prop->{ID_FS_TYPE}};
    $dc = $dumpAndCheck{$clas};

    # Filesystems nativos (se pueden checkear al arranque con fstab)
    return $opciones{'nativo'} . "\t" . $dc if ($clas eq "nativo");
    
    return $opciones{'swap'} . "\t" . $dumpAndCheck{'swap'} if ($clas eq "swap");
    
    # Si llega hasta acá no es nativo
    return $opciones{$prop->{'ID_FS_TYPE'}} . "\t" . $dumpAndCheck{'no_nativo'};
}


#Agregamos un grupo que pueda escribir en las particiones vfat
system("addgroup --gid 500 particiones > /dev/null 2>&1");
#El installer no ejecuta udevd dentro del chroot (requerido por udevadm)
if (! -e "/var/run/udev" && ! -e "/dev/.udev"){
	system("udevd &");
}

#system("mount /proc");
#system("mount -t sysfs sysfs /sys");

sub es_removible($){
# Recibe el path de un dispositivo de bloques (o de una partición de un dispositivo)
# y retorna true si es un dispositivo removible (como un pendrive)
    my $device = $_[0];
    my $removable = "";
    $device =~ s%/dev/([^0-9]+)[0-9]*%$1%;
    open(REMOVABLE, "</sys/block/$device/removable");
    $removable = <REMOVABLE>;
    close(REMOVABLE);
    return $removable;
}
		  
sub es_floppy($){
#	print "Floppy: " . defined($_[0]->{"ID_DRIVE_FLOPPY"}) . "\n";
	return defined($_[0]->{"ID_DRIVE_FLOPPY"});
}

sub es_cdrom($){
	return defined($_[0]->{"ID_TYPE"}) && $_[0]->{"ID_TYPE"} eq "cd";
}

sub textArrayToHash($){
# Convierte una lista de strings de la forma "CLAVE=VALOR" a un hash, retorna
# una referencia a ese hash
    my $array = $_[0];
    my $hashTable = {};
    foreach my $field (@$array){
        my ($key, $value) = split("=", $field);
		chomp $value;
        $hashTable->{$key} = $value;
    }
    return $hashTable;
}
sub info_disco($){
# Utiliza udevadm para retornar hash con las propiedades del disco
# entre ellas tipo de disco, filesystem, links que apuntan al dispositivo y
# bus
    my $nombre = $_[0];
    my @info_disco;
    system("/sbin/udevadm trigger");
    open UDEVADM, "/sbin/udevadm info --query=property --name=$nombre |";
    @info_disco = <UDEVADM>;
    close UDEVADM;
    return textArrayToHash(\@info_disco);
}

sub detectar_discos(){
# Detecta los discos y genera un hash donde la clave es el nombre del dispositivo
# y el valor es una referencia a un hash con sus propiedades
	my $info_discos = {};
	my $path = "/sys/block";
	opendir BLOCK, $path;
	my @discos = readdir BLOCK;
	foreach my $disco (@discos){
		if ( -d $disco ){
			next; # Se saltea "." y "..", el resto son links simbólicos
		}

		my $info_prop = info_disco($disco);
		# Ignoramos los USB y los que no sean discos
		if ($info_prop->{"DEVTYPE"} ne "disk" || (defined($info_prop->{"ID_BUS"}) && $info_prop->{"ID_BUS"} eq "usb")){
			next;
		}
		# Agregamos un campo más al hash
		$info_prop->{"X-REMOVABLE"} = es_removible($disco);
		
		$info_discos->{$disco} = $info_prop;
	}
	close BLOCK;
	return $info_discos;
}

sub agregar_a_fstab($$){
# Imprime una entrada para un disco, identifica al disco usando su
# label, uuid o dispositivo (dependiendo de lo que tenga configurado,
# en ese orden)
	my ($prop, $mountpoint) = @_;
	my ($id, $fs);

	if (fs_previamente_configurado($prop) ){
		return "";
	}
	descripcion_fs($prop, $mountpoint);
	if (defined($prop->{"ID_FS_UUID"})){
		$id = "UUID=" . $prop->{"ID_FS_UUID"};
	}
	else{
		$id = $prop->{"DEVNAME"};
	}
	$prop->{'ID_FS_TYPE'} =~ s/^ntfs$/ntfs-3g/;
	mkdir($mountpoint) if ($mountpoint =~ /^\//); # Evitamos los path relativos, por ej. "none"
	print "$id\t$mountpoint\t$prop->{'ID_FS_TYPE'}\t" . opciones($prop, 0) . "\n\n";
}
sub agregar_removible_a_fstab($$){
# Imprime una entrada para un floppy o un cd/dvd
	my ($prop, $mountpoint) = @_;
	my $fs;
	if (fs_previamente_configurado($prop)){
		return "";
	}
    # Si es un disquete (asumimos vfat)
	if (es_floppy($prop)){
		$fs = 'auto';
	}   
	elsif (es_cdrom($prop)){
		$fs = 'udf,iso9660';
	}   
	mkdir($mountpoint);
	print "$prop->{'DEVNAME'}\t$mountpoint\t$fs\t" . opciones($prop, 0) . "\n\n";
}

sub descripcion_fs($$){
	my ($prop, $mountpoint) = @_;
	print("# $mountpoint se encontraba en $prop->{'DEVNAME'} al momento de la instalación\n");
}

sub fs_en_fstab{
# Retorna un hash con los filesystems presentes en /etc/fstab
# las claves del hash son la ruta al dispositivo (/dev/hda1 por ejemplo)
# también pueden ser UUIDs o LABELs
# y el valor para cada clave es la lista de palabras de la línea
# completa de /etc/fstab para ese dispositivo
	my (%particiones, @linea, $fstab);
	open($fstab, "</etc/fstab");
	while (<$fstab>){
		if (/^\s*[^#\s]/){
			@linea = split(" ");
			$particiones{$linea[0]} = [@linea];
		}
	}
	close($fstab);
	return %particiones;
}

my %fstab_hash = fs_en_fstab();

sub fs_previamente_configurado($){
# Retorna true si el filesystem ya está en el fstab
	my $prop = $_[0];
	return (defined($prop->{"ID_FS_LABEL"}) && defined($fstab_hash{"LABEL=".$prop->{'ID_FS_LABEL'}})) ||
	(defined($prop->{"ID_FS_UUID"}) && defined($fstab_hash{"UUID=".$prop->{'ID_FS_UUID'}})) ||
	defined($fstab_hash{$prop->{"DEVNAME"}});
}

print("\n## Configuración agregada por lihuenconfig ##\n");

my %discos = %{ detectar_discos() };
my $info_prop;
my ($num_disco, $num_disq, $num_cd) = (1, 0, 0);
while ( my ($nombre, $propiedades) = each %discos ){
#agregar_a_fstab($info_prop);
	if (es_floppy($propiedades)){
		agregar_removible_a_fstab($propiedades, "/media/disquetera${num_disq}");
		$num_disq++;
	}
	elsif (es_cdrom($propiedades)){
		agregar_removible_a_fstab($propiedades, "/media/cdrom${num_cd}");
		$num_cd++;
	}
	elsif (defined($propiedades->{"ID_PART_TABLE_TYPE"})){
		for (my $i = 1; $i <= 99; $i++){
			if ($i > 4 && ! -e $propiedades->{"DEVNAME"}.$i){
				last;
			}
			elsif ( ! -e $propiedades->{"DEVNAME"}.$i ){
				next;
			}
			my $particion = $propiedades->{"DEVNAME"}.$i;
			$particion =~ s/$basename/$1/g;
			$info_prop = info_disco($particion);
			$info_prop->{"X-REMOVABLE"} = $propiedades->{"X-REMOVABLE"};
			if (defined($info_prop->{"ID_FS_TYPE"})){
				if ($info_prop->{"ID_FS_TYPE"} eq "swap"){
					agregar_a_fstab($info_prop, "none");
				}
				else{
					agregar_a_fstab($info_prop, "/mnt/disco${num_disco}part${i}");
				}
			}
		}
		$num_disco++;
	}
	elsif (defined($propiedades->{"ID_FS_TYPE"})){
		agregar_a_fstab($propiedades, "/mnt/disco${num_disco}");
		$num_disco++;
	}
}
